%* cst_createds                                                                   *;
%*                                                                                *;
%* Creates a zero-observation data set that is based on a registered standard.    *;
%*                                                                                *;
%* Note:  Use of this macro is recommended only in SAS Clinical Standards Toolkit *;
%*        1.4 and earlier. Equivalent and updated functionality is provided with  *;
%*        cst_createdsfromtemplate() for Toolkit 1.5 and later.                   *;
%*                                                                                *;
%* @deprecated Replaced, effective with CST 1.5, with cst_createdsfromtemplate    *;
%*                                                                                *;
%* @macvar _cstSeqCnt Results: Sequence number within _cstResultSeq               *;
%* @macvar _cstResultSeq Results: Unique invocation of check                      *;
%* @macvar _cstMessages  Cross-standard work messages data set                    *;
%* @macvar _cst_rc Task error status                                              *;
%* @macvar _cstLRECL Logical record length setting for filename statement         *;
%*                                                                                *;
%* @param  _cstStandard - required - The name of a registered standard.           *;
%* @param  _cstStandardVersion - optional - The version of the standard that the  *;
%*             data set should be created from. If this is omitted, the default   *;
%*             version for the standard is used. If a default version is not      *;
%*             defined, an error is generated.                                    *;
%* @param  _cstType - required -The type of data set to create. This value comes  *;
%*             from the TYPE column in the SASReferences file for the standard-   *;
%*             version combination.                                               *;
%* @param  _cstSubType - optional - The subtype for the type. This value comes    *;
%*             from the SUBTYPE column in the SASReferences file for the          *;
%*             standard-version combination. If the type has no subtypes, the     *;
%*             value can be omitted. Otherwise, it must be provided.              *;
%* @param  _cstOutputDS - required - The name of the data set to create.          *;
%* @param  _cstOutputDSLabel - optional - The label of the data set to create.    *;
%* @param  _cstResultsOverrideDS - optional - The (libname.)member that refers to *;
%*             the Results data set to create. If omitted, the Results data set   *;
%*             that is specified by &_cstResultsDS is used.                       *;
%*                                                                                *;
%* @since 1.2                                                                     *;
%* @exposure  external                                                            *;

%macro cst_createDS(
    _cstStandard=,
    _cstStandardVersion=,
    _cstType=,
    _cstSubType=,
    _cstOutputDS=,
    _cstOutputDSLabel=,
    _cstResultsOverrideDS=
    ) / des='CST: Creates a zero observation data set';

  %cstutil_setcstgroot;

  %* declare local variables used in the macro;
  %local
    _cstError
    _cstGlobalMDLib
    _cstGlobalMDPath
    _cstGlobalStdDS
    _cstGlobalStdSASRefsDS
    _cstMsgDir
    _cstMsgMem
    _cstNeedToDeleteMsgs
    _cstNextCode
    _cstParm1
    _cstParm2
    _cstParamInError
    _cstRandom
    _cstSaveResultSeq
    _cstSaveSeqCnt
    _cstTempDS1
    _cstTempLib2
    _cstThisMacroRC
    _cstThisResultsDS
    _cstThisResultsDSLib
    _cstThisResultsDSMem
    _cstUsingResultsOverride
    ;

  %let _cstThisMacroRC=0;

  %* retrieve static variables;
  %cst_getStatic(_cstName=CST_GLOBALMD_PATH,_cstVar=_cstGlobalMDPath);
  %cst_getStatic(_cstName=CST_GLOBALMD_REGSTANDARD,_cstVar=_cstGlobalStdDS);
  %cst_getStatic(_cstName=CST_GLOBALMD_SASREFS,_cstVar=_cstGlobalStdSASRefsDS);

  %* Generate the random names used in the macro;
  %cstutil_getRandomNumber(_cstVarname=_cstRandom);
  %let _cstGlobalMDLib=_cst&_cstRandom;

  %cstutil_getRandomNumber(_cstVarname=_cstRandom);
  %let _cstTempDS1=_cst&_cstRandom;

  %cstutil_getRandomNumber(_cstVarname=_cstRandom);
  %let _cstTempLib2=_cst&_cstRandom;

  %cstutil_getRandomNumber(_cstVarname=_cstRandom);
  %let _cstNextCode=_cst&_cstRandom;

  %* Create a temporary messages data set if required;
  %cstutil_createTempMessages(_cstCreationFlag=_cstNeedToDeleteMsgs);

  %* Set up the location of the results data set to write to;
  %cstutil_internalManageResults(_cstAction=SAVE);
  %let _cstResultSeq=1;
  %let _cstSeqCnt=0;

  * Assign the libname to the global metadata library;
  libname &_cstGlobalMDLib "%unquote(&_cstGlobalMDPath)" access=readonly;

  %* Pre-requisite: _cstStandard is not blank;
  %if (%length(&_cstStandard)=0) %then %do;
    %* process an error;
    %let _cstParamInError=_cstStandard;
    %goto NULL_PARAMETER;
  %end;

  %* Pre-requisite: _cstType is not blank;
  %if (%length(&_cstType)=0) %then %do;
    %let _cstParamInError=_cstType;
    %goto NULL_PARAMETER;
  %end;

  %* Pre-requisite: _cstOutputDS is not blank;
  %if (%length(&_cstOutputDS)=0) %then %do;
    %let _cstParamInError=_cstOutputDS;
    %goto NULL_PARAMETER;
  %end;

  * Pre-requisite: Check that the standard is valid;
  %let _cstError=Y;
  data &_cstTempDS1;
    set &_cstGlobalMDLib..&_cstGlobalStdDS
      (where=(upcase(standard)=%sysfunc(upcase("&_cstStandard"))));
    if (_n_=1) then call symputx('_cstError','N','L');
  run;

  %if (&_cstError=Y) %then %do;
    %goto INVALID_STD;
  %end;

  * Pre-requisite: Check that the version is valid (if provided) or a default exists if not;
  %if (%length(&_cstStandardVersion)>0) %then %do;
    %let _cstError=Y;
    * Check that the version is valid;
    data &_cstTempDS1;
      set &_cstTempDS1
        (where=(upcase(standardversion)=%sysfunc(upcase("&_cstStandardVersion"))));
      if (_n_=1) then call symputx('_cstError','N','L');
    run;

    %* process an error;
    %if (&_cstError=Y) %then %do;
      %goto INVALID_VERSION;
    %end;
  %end;
  %else %do;
    %let _cstError=Y;
    * Check that there is a default version specified and retrieve it;
    data &_cstTempDS1;
      set &_cstTempDS1
        (where=(isstandarddefault="Y"));
      if (_n_=1) then call symputx('_cstError','N','L');
      call symputx('_cstStandardVersion',standardVersion,'L');
    run;

    %* process an error;
    %if (&_cstError=Y) %then %do;
      %let _cstParm1=&_cstStandardVersion;
      %let _cstParm2=&_cstStandard;
      %goto NO_DEFAULT_VERSION;
    %end;
  %end;

  * Pre-requisite: Check that the type is valid;
  %let _cstError=Y;
  data &_cstTempDS1;
    set &_cstGlobalMDLib..&_cstGlobalStdSASRefsDS
      (where=(
        upcase(standard)=%sysfunc(upcase("&_cstStandard")) and
        upcase(standardVersion)=%sysfunc(upcase("&_cstStandardVersion")) and
        upcase(type)=%sysfunc(upcase("&_cstType"))
      ));
    if (_n_=1) then call symputx('_cstError','N','L');
  run;

  %if (&_cstError=Y) %then %do;
    %let _cstParm1=&_cstType;
    %let _cstParm2=&_cstStandard &_cstStandardVersion;
    %goto INVALID_TYPE;
  %end;

  * Pre-requisite: Validate subtype;
  %if (%length(&_cstSubType)=0) %then %do;
    * Check that there is only one matching record;
    %let _cstError=N;
    data _null_;
      set &_cstTempDS1;
      if (_n_ > 1) then do;
        call symputx('_cstError','Y','L');
        stop;
      end;
    run;

    %if (&_cstError=Y) %then %do;
      %let _cstParm1=&_cstType;
      %goto MULTIPLE_TYPE_RECS;
    %end;
  %end;
  %else %do;
    %let _cstError=Y;
    data &_cstTempDS1;
      set &_cstTempDS1
        (where=(
          upcase(subtype)=%sysfunc(upcase("&_cstSubtype"))
        ));
      if (_n_=1) then call symputx('_cstError','N','L');
    run;

    %if (&_cstError=Y) %then %do;
      %let _cstParm1=&_cstType;
      %if (%length(&_cstSubType)>0) %then %do;
        %let _cstParm1= &_cstType./&_cstSubType;
      %end;
      %let _cstParm2=&_cstStandard &_cstStandardVersion;
      %goto INVALID_SUBTYPE;
    %end;
  %end;

  %let _cstError=N;
  data &_cstTempDS1;
    set &_cstTempDS1;
      if (memname='') then call symputx('_cstError','Y','L');
  run;

  %if (&_cstError=Y) %then %do;
    %let _cstParm1=A nonmissing value for memname for  type &_cstType;
    %if (%length(&_cstSubType)>0) %then %do;
      %let _cstParm1= A nonmissing value for memname for  type/subtype &_cstType./&_cstSubType;
    %end;
    %goto INVALID_MEMNAME;
  %end;

  * Assign a filename for the code that will be generated;
  filename &_cstNextCode CATALOG "work.&_cstNextCode..nextcode.source" &_cstLRECL;

  data _null_;
    file &_cstNextCode;
    set &_cstTempDS1;

      attrib _cstTempVar _cst14_keeplist format=$500.;

      * handle the case where membername has been given an extension;
      memname=scan(memname,1,'.');

      * Make changes here specifically to accommodate CST 1.5 column additions to SASReferences.  *;
      * This provides backward compatibility for < CST 1.5 code that uses this macro.             *;
      if upcase(subtype) in ("REFERENCE" "SASREFERENCES") then
      do;
        _cst14_keeplist='keep=standard standardversion type subtype SASref reftype path order memname comment';
      end;

      put 'libname &_cstNextCode "' %unquote(path) +(-1)'" access=readonly;';
      put 'proc sql;';

      _cstTempVar = catx(' ',cats("create table &_cstOutputDS (",strip(_cst14_keeplist)),cats("label=",'"',
                                  "&_cstOutputDSLabel",'") like'),cats("&_cstNextCode..",memname),";");
      put @3 _cstTempVar;
      put @2 'quit;';

      put "libname &_cstNextCode ;";
  run;

  * Include the generated code;
  %include &_cstNextCode;

  * Clear the filename;
  filename &_cstNextCode;

  %let _cstThisMacroRC=0;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0102
                ,_cstResultParm1=&_cstOutputDS
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );
  %goto CLEANUP;

%NULL_PARAMETER:
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0081
                ,_cstResultParm1=&_cstParamInError
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );
  %goto CLEANUP;

%INVALID_STD:
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0082
                ,_cstResultParm1=&_cstStandard
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );

  %goto CLEANUP;

%INVALID_VERSION:
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0083
                ,_cstResultParm1=&_cstStandardVersion
                ,_cstResultParm2=&_cstStandard
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );

  %goto CLEANUP;

%INVALID_TYPE:
  %* ERROR: The type &_cstType is not defined for &_cstStandard &_cstStandardVersion..;
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0084
                ,_cstResultParm1=&_cstParm1
                ,_cstResultParm2=&_cstParm2
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );
  %goto CLEANUP;

%NO_DEFAULT_VERSION:
  %* ERROR: No version was supplied and there is no default for &_cstStandard.;
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0083
                ,_cstResultParm1=&_cstParm1
                ,_cstResultParm2=&_cstParm2
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );

  %goto CLEANUP;

%MULTIPLE_TYPE_RECS:
  %* ERROR: The type &_cstType has more than one subtype and none was specified.;
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0086
                ,_cstResultParm1=&_cstParm1
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );

  %goto CLEANUP;

%INVALID_SUBTYPE:
  %* ERROR: The type/subtype &_cstType./&_cstSubType is not defined for &_cstStandard &_cstStandardVersion..;
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0087
                ,_cstResultParm1=&_cstParm1
                ,_cstResultParm2=&_cstParm2
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );

  %goto CLEANUP;

%INVALID_MEMNAME:
  %* ERROR: No data set memname is available for the specified type/subtype;
  %let _cstThisMacroRC=1;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
                _cstResultId=CST0008
                ,_cstResultParm1=&_cstParm1
                ,_cstResultParm2=
                ,_cstResultSeqParm=&_cstResultSeq
                ,_cstSeqNoParm=&_cstSeqCnt
                ,_cstSrcDataParm=CST_CREATEDS
                ,_cstResultFlagParm=&_cstThisMacroRC
                ,_cstRCParm=&_cstThisMacroRC
                ,_cstResultsDSParm=&_cstThisResultsDS
                );

  %goto CLEANUP;

%CLEANUP:
  %* reset the resultSequence/SeqCnt variables;
  %cstutil_internalManageResults(_cstAction=RESTORE);

  * Clear the libname;
  libname &_cstGlobalMDLib;

  * Clean up temporary data sets if they exist;
  proc datasets nolist lib=work;
    delete &_cstTempDS1  / mt=data;
    delete &_cstNextCode / mt=catalog;
    quit;
  run;

  %* Delete the temporary messages data set if it was created here;
  %if (&_cstNeedToDeleteMsgs=1) %then %do;
    %if %eval(%index(&_cstMessages,.)>0) %then %do;
      %let _cstMsgDir=%scan(&_cstMessages,1,.);
      %let _cstMsgMem=%scan(&_cstMessages,2,.);
    %end;
    %else %do;
      %let _cstMsgDir=work;
      %let _cstMsgMem=&_cstMessages;
    %end;
    proc datasets nolist lib=&_cstMsgDir;
      delete &_cstMsgMem / mt=data;
      quit;
    run;
  %end;

  %let _cst_rc=&_cstThisMacroRC;
%mend cst_createDS;
