%* sendutil_createsrcmetafromsaslib                                               *;
%*                                                                                *;
%* Derive source metadata files from a SAS data library for a CDISC SEND study.   *;
%*                                                                                *;
%* This sample utility macro attempts to derive source metadata files  from a SAS *;
%* data library for a CDISC SEND study.                                           *;
%*                                                                                *;
%* These source metadata files are used by the SAS Clinical Standards Toolkit to  *;
%* support CDISC SEND validation and to derive CDISC CRT-DDS (define.xml) files:  *;
%*          source_tables                                                         *;
%*          source_columns                                                        *;
%*          source_study                                                          *;
%*                                                                                *;
%* This is the general strategy:                                                  *;
%*    1. Use PROC CONTENTS output as the primary source of the information.       *;
%*    2. Use reference_tables and reference_columns for matching the columns.     *;
%*    3. Use class_columns as a generic source for the metadata.                  *;
%*                                                                                *;
%* NOTE:  This is ONLY an attempted approximation of source metadata. No          *;
%*        assumptions should be made that the result accurately represents the    *;
%*        study data.                                                             *;
%*                                                                                *;
%* ASSUMPTIONS:                                                                   *;
%*    1. The source data is read from a single SAS library. You can modify the    *;
%*       code to reference multiple libraries by using library concatenation.     *;
%*    2. The data set keys are estimated by the sort order of the source data     *;
%*       (if specified). If it is not specified, the data set keys are estimated  *;
%*       based on columns that SAS uses to define keys in the reference standard. *;
%*    3. For any unknown domain, the domain class (Events, Interventions, or      *;
%*       Findings) is estimated based on the class-specific topic variable (that  *;
%*       is, _TERM (events), _TRT (interventions), and _TESTCD (findings)).       *;
%*    4. Most column values in source_study are hardcoded because there is no     *;
%*       metadata source. These values are used only to build the define.xml file.*;
%*       These values are marked as  <--- HARDCODE  below.                        *;
%*                                                                                *;
%* LIMITATIONS:                                                                   *;
%*   1. Here is a scenario that has not been addressed:                           *;
%*       - Split domains, such as QS**                                            *;
%*                                                                                *;
%* @macvar studyRootPath Root path to the sample source study                     *;
%* @macvar _cstClassColumnDS Reference standard class columns metadata data set   *;
%* @macvar _cstDebug Turns debugging on or off for the session                    *;
%* @macvar _cstGRoot Root path of the Toolkit Global Library                      *;
%* @macvar _cstResultsDS Results data set                                         *;
%* @macvar _cstSASRefs Run-time SASReferences data set derived in process setup   *;
%* @macvar _cstSENDDataLib SEND source data library                               *;
%* @macvar _cstSeqCnt Results: Sequence number within _cstResultSeq               *;
%* @macvar _cstSrcData Results: Source entity being evaluated                     *;
%* @macvar _cstStandard Name of a standard registered to Toolkit                  *;
%* @macvar _cstStandardVersion Version of the standard referenced in _cstStandard *;
%* @macvar _cstTrgStudyDS Derived source_study data set                           *;
%*                                                                                *;
%* @since 1.5                                                                     *;
%* @exposure external                                                             *;

%macro sendutil_createsrcmetafromsaslib(
    ) / des='CST: Create SEND metadata from SAS library';

  %local
    _cstDataRecords
    _cstrundt
    _cstVer
  ;

  %let _cstSeqCnt=0;
  %let _cstSrcData=&sysmacroname;


  %* Write information about this process to the results data set  *;
  %if %symexist(_cstResultsDS) %then
  %do;
    %cst_getRegisteredStandards(_cstOutputDS=work._cstStandards);

    data _null_;
      set work._cstStandards (where=(standard="CST-FRAMEWORK"));
      call symputx('_cstVer',strip(productrevision));
      call symputx('_cstrundt',put(datetime(),is8601dt.));
    run;
    proc datasets lib=work nolist;
      delete _cstStandards;
    quit;

    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STANDARD: &_cstStandard,_cstSeqNoParm=1,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STANDARDVERSION: &_cstStandardVersion,_cstSeqNoParm=2,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS DRIVER: CREATE_SOURCEMETADATA,_cstSeqNoParm=3,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS DATE: &_cstrundt,_cstSeqNoParm=4,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS TYPE: METADATA DERIVATION,_cstSeqNoParm=5,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS SASREFERENCES: &_cstsasrefs,_cstSeqNoParm=6,_cstSrcDataParm=&_cstSrcData);
    %if %symexist(studyRootPath) %then
      %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STUDYROOTPATH: &studyRootPath,_cstSeqNoParm=7,_cstSrcDataParm=&_cstSrcData);
    %else
      %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STUDYROOTPATH: <not used>,_cstSeqNoParm=7,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS GLOBALLIBRARY: &_cstGRoot,_cstSeqNoParm=8,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS CSTVERSION: &_cstVer,_cstSeqNoParm=9,_cstSrcDataParm=&_cstSrcData);
    %let _cstSeqCnt=9;
  %end;


  * A single source data library serves as the input to this process.  *;
  proc contents data=&_cstCRTDataLib.._all_ out=work.contents
      (keep=memname memlabel name type length label varnum formatl formatd sorted sortedby)  noprint;
  run;

  ***************************************;
  * Begin derivation of source_study    *;
  ***************************************;

  * Create a sample source_study data set that serves as input  *;
  *  to build a define file from SEND source metadata.          *;
  data work.source_study;
    attrib
      definedocumentname format=$64. label="Unique study ID"
      sasref format=$8. label="SAS metadata libref"
      studyname format=$128. label="Study name"
      protocolname format=$128. label="Protocol name"
      studydescription format=$2000. label="Description of the study"
    ;

    definedocumentname='SAS_CST_Define';                                                        * <--- HARDCODE  *;
    sasref=upcase("&_cstSENDDataLib");
    studyname='Derived Study built by SAS Clinical Standards Toolkit';                          * <--- HARDCODE  *;
    protocolname='SAS_CST_Define Sample Protocol';                                              * <--- HARDCODE  *;
    studydescription=catx(' ','Derived Study built from data in',pathname("&_cstCRTDataLib"));  * <--- HARDCODE  *;
    output;
  run;

* Write out final study-level source metadata  *;
proc sort data=work.source_study out=&_cstTrgMetaLibrary..&_cstTrgStudyDS (label="Source Study Metadata");
  by definedocumentname studyname sasref;
run;

%if %symexist(_cstResultsDS) %then
%do;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
      _cstResultID=CST0074,
      _cstResultParm1=%sysfunc(pathname(&_cstTrgMetaLibrary)),
      _cstSeqNoParm=&_cstSeqCnt,
      _cstSrcdataParm=&_cstSrcData,
      _cstActualParm=&_cstTrgStudyDS
);
%end;

***************************************;
* Begin derivation of source_tables   *;
***************************************;

* This section attempts to guess information about unrecognized domains (custom or SUPP*)  *;
*  based on available SAS metadata and columns contained within each domain.               *;

proc sort data=work.contents;
  by memname sortedby;
run;

data work.tables (drop=name sorted sortedby tempdomain tempvar dsid rc);
  set work.contents (keep=memname memlabel name sorted sortedby);
    by memname;
  attrib tempkeys format=$200.
         tempclass format=$40.
         tempdomain format=$20.
         tempvar format=$8.;
  retain tempkeys tempclass;
  if first.memname then
  do;
    tempkeys='';
    tempclass='';
  end;

  * First look to see if the data set is sorted, and if so assume the sort columns as keys *;
  if sorted=1 then
  do;
    if sortedby ne . then
      tempkeys = catx(' ',tempkeys,name);
  end;
  * Otherwise, estimate the class and keys based on the data set columns  *;
  else
  do;
    if (substr(memname,1,4))='SUPP' then
    do;
       if tempclass = '' then do;
         tempclass='Relates';
         tempkeys='STUDYID RDOMAIN USUBJID IDVAR IDVARVAL QNAM';
       end;
    end;
    else do;
      * Interventions topic variable:  **TRT  *;
      if substr(name,3)='TRT' then
      do;
        tempclass='Interventions';
        tempkeys=catx(' ','STUDYID USUBJID',name,cats(memname,'STDTC'));
        tempdomain = catx('.',upcase("&_cstCRTDataLib"),memname);
        dsid=open(tempdomain);
        if dsid ne 0 then
          if varnum(dsid,cats(memname,'STDTC'))=0 then tempkeys=TRANWRD(tempkeys,cats(memname,'STDTC'),'');
        tempkeys = compbl(tempkeys);
        rc=close(dsid);
      end;
      * Findings topic variable:  **TESTCD  *;
      else if substr(name,3)='TESTCD' then
      do;
        tempclass='Findings';
        tempkeys=catx(' ','STUDYID USUBJID',name,'VISITNUM',cats(memname,'TPTREF'),cats(memname,'TPTNUM'));
        tempdomain = catx('.',upcase("&_cstCRTDataLib"),memname);
        dsid=open(tempdomain);
        if dsid ne 0 then
          do tempvar = 'VISITNUM',cats(memname,'TPTREF'),cats(memname,'TPTNUM');
            if varnum(dsid,tempvar)=0 then tempkeys=TRANWRD(tempkeys,tempvar,'');
          end;
          tempkeys = compbl(tempkeys);
          rc=close(dsid);
        end;
      * Events topic variable:  **TERM  *;
      else if substr(name,3)='TERM' then
      do;
        tempclass='Events';
        tempkeys=catx(' ','STUDYID USUBJID',cats(memname,'DECOD'),cats(memname,'STDTC'));
        tempdomain = catx('.',upcase("&_cstCRTDataLib"),memname);
        dsid=open(tempdomain);
        if dsid ne 0 then
          do tempvar = cats(memname,'DECOD'),cats(memname,'STDTC');
            if varnum(dsid,tempvar)=0 then tempkeys=TRANWRD(tempkeys,tempvar,'');
          end;
          tempkeys = compbl(tempkeys);
          rc=close(dsid);
      end;
    end;
  end;
  if last.memname then output;
run;

* Split processing for domains found in the reference standard (work.ref_tables)    *;
*  and those unrecognized (custom and SUPP*) domains (work.new_tables).             *;
proc sql noprint;
  create table work.ref_tables as
  select upcase("&_cstSENDDataLib") as sasref length=8 format=$8. label='SASreferences sourcedata libref',
         ref.table,
         case when memlabel ne '' then memlabel
              else label
         end as label length=40 format=$40. label='Table Label',
         ref.class,
         ref.xmlpath,
         ref.xmltitle,
         ref.structure,
         ref.purpose,
         ref.keys,
         ref.state,
         ref.date,
         ref.standard,
         ref.standardversion,
         ref.standardref,
         ref.comment from
    &_cstRefLib..&_cstRefTableDS ref
      full join
    work.tables
    on ref.table=tables.memname
    where tables.memname ne '' and ref.table ne '' and substr(tables.memname,1,4) ne 'SUPP'
    order by table;
  create table work.new_tables as
  select upcase("&_cstSENDDataLib") as sasref length=8 format=$8. label='SASreferences sourcedata libref',
         case when memname ne '' then memname
              else table
         end as table,
         case when memlabel ne '' then memlabel
              else label
         end as label length=40 format=$40. label='Table Label',
         case when tempclass ne '' then tempclass
              else class
         end as class,
         cats('../',lowcase(calculated table),'.xpt') as xmlpath,
         catx(' ',calculated table,calculated label,'SAS transport file') as xmltitle,
         ref.structure,
         ref.purpose,
         case when tempkeys ne '' then tempkeys
              else keys
         end as keys,
         ref.state,
         ref.date,
         ref.standard,
         ref.standardversion,
         ref.standardref,
         ref.comment from
    &_cstRefLib..&_cstRefTableDS ref
      full join
    work.tables
    on ref.table=tables.memname
    where (tables.memname ne '' and ref.table eq '') or substr(tables.memname,1,4) eq 'SUPP'
    order by table;
quit;

* Put all the tables back together again.  *;
data work.source_tables;
  set work.ref_tables
      work.new_tables;
run;

* Write out final table-level source metadata  *;
proc sort data=work.source_tables out=&_cstTrgMetaLibrary..&_cstTrgTableDS  (label="Source Table Metadata");
  by sasref table;
run;

%if %symexist(_cstResultsDS) %then
%do;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
      _cstResultID=CST0074,
      _cstResultParm1=%sysfunc(pathname(&_cstTrgMetaLibrary)),
      _cstSeqNoParm=&_cstSeqCnt,
      _cstSrcdataParm=&_cstSrcData,
      _cstActualParm=&_cstTrgTableDS
);
%end;

***************************************;
* Begin derivation of source_columns  *;
***************************************;

data work.columns;
  set work.contents (drop=memlabel sorted sortedby rename=(memname=table type=ctype label=clabel length=clength));
run;

* Split processing for columns found in the reference standard (work.ref_columns)   *;
*  and those unrecognized (custom and SUPP*) columns (work.new_columns).            *;
proc sql noprint;
  create table work.ref_columns as
  select upcase("&_cstSENDDataLib") as sasref length=8 format=$8. label='SASreferences sourcedata libref',
         ref.table,
         ref.column,
         case when clabel ne '' then clabel
              else label
         end as label length=200 format=$200. label='Column Description',
         columns.varnum as order format=8. label='Column Order',
         case when ctype=1 then 'N'
              else 'C'
         end as type format=$1. label='Column Type',
         columns.clength as length format=8. label='Column Length',
         case when formatl>0 then cats(formatl,'.',formatd)
              else ref.displayformat
         end as displayformat,
         ref.xmldatatype,
         ref.xmlcodelist,
         ref.core,
         ref.origin,
         ref.role,
         ref.term,
         ref.algorithm,
         ref.qualifiers,
         ref.standard,
         ref.standardversion,
         ref.standardref,
         ref.comment,
         1 as _cstfound from
    &_cstRefLib..&_cstRefColumnDS ref
      full join
    work.columns
    on ref.table=columns.table and ref.column=columns.name
    where columns.name ne '' and ref.column ne ''
    order by table, column;
  create table work.new_columns as
  select upcase("&_cstSENDDataLib") as sasref length=8 format=$8. label='SASreferences sourcedata libref',
         columns.table,
         columns.name as column,
         case when clabel ne '' then clabel
              else label
         end as label length=200 format=$200. label='Column Description',
         columns.varnum as order format=8. label='Column Order',
         case when ctype=1 then 'N'
              else 'C'
         end as type format=$1. label='Column Type',
         columns.clength as length format=8. label='Column Length',
         ref.displayformat,
         ref.xmldatatype,
         ref.xmlcodelist,
         ref.core,
         ref.origin,
         ref.role,
         ref.term,
         ref.algorithm,
         ref.qualifiers,
         ref.standard,
         ref.standardversion,
         ref.standardref,
         ref.comment from
    &_cstRefLib..&_cstRefColumnDS ref
      full join
    work.columns
    on ref.table=columns.table and ref.column=columns.name
    where columns.name ne '' and ref.column eq ''
    order by table, column;
quit;

* Add class to each record *;
proc sql noprint;
  create table work.new_classcolumns as
  select tab.class, col.* from
  work.new_columns col
   left join
  work.source_tables tab
  on col.table = tab.table;
quit;

* Step 1:  Now go to refmeta.class_columns to get some generic metadata on any of the new columns    *;
*  This step will only find metadata for columns in the Findings, Events and Interventions classes.  *;
proc sql noprint;
  create table work.new_columns_step1 as
  select new.sasref,
         new.table,
         new.column,
         case when new.label ne '' then new.label
              else class.label
         end as label,
         new.order,
         new.type,
         new.length,
         class.displayformat,
         class.xmldatatype,
         class.xmlcodelist,
         class.core,
         class.origin,
         class.role,
         class.term,
         class.algorithm,
         class.qualifiers,
         class.standard,
         class.standardversion,
         class.standardref,
         class.comment,
         case when class.column ne '' then 1
              else 0
         end as _cstfound
 from
  work.new_classcolumns new
   left join
  &_cstRefLib..&_cstClassColumnDS class
  on upcase(new.class) = class.table and substr(new.column,3) = substr(class.column,3)
  order by table, column ;
quit;

* This step does a lookup for all domain identifier and timing columns  *;
*  (those without the 2-character domain prefix).                       *;
proc sql noprint;
  create table work.new_columns_step2 as
  select new.sasref,
         new.table,
         new.column,
         case when new.label ne '' then new.label
              else class.label
         end as label,
         new.order,
         new.type,
         new.length,
         class.displayformat,
         class.xmldatatype,
         class.xmlcodelist,
         class.core,
         class.origin,
         class.role,
         class.term,
         class.algorithm,
         class.qualifiers,
         class.standard,
         class.standardversion,
         class.standardref,
         class.comment,
         case when class.column ne '' then 1
              else 0
         end as _cstfound
 from
  work.new_columns_step1 (where=(_cstfound=0)) new
   left join
  &_cstRefLib..&_cstClassColumnDS class
  on new.column = class.column
  order by table, column ;
quit;

* This step does a lookup for all non-identifier domain specific columns  *;
*  (those with the 2-character domain prefix).                            *;
proc sql noprint;
  create table work.new_columns_step3 as
  select new.sasref,
         new.table,
         new.column,
         case when new.label ne '' then new.label
              else class.label
         end as label,
         new.order,
         new.type,
         new.length,
         class.displayformat,
         class.xmldatatype,
         class.xmlcodelist,
         class.core,
         class.origin,
         class.role,
         class.term,
         class.algorithm,
         class.qualifiers,
         class.standard,
         class.standardversion,
         class.standardref,
         class.comment,
         case when class.column ne '' then 1
              else 0
         end as _cstfound
 from
  work.new_columns_step2 (where=(_cstfound=0)) new
   left join
  &_cstRefLib..&_cstClassColumnDS class
  on substr(new.column,3) = substr(class.column,3)
  order by table, column ;
quit;

* This step does a lookup for any remaining Supplemental Qualifier columns  *;
* Note the code assumes a single SUPPQUAL reference data set.  If the       *;
*  reference table data set contains multiple SUPP* domain definitions,     *;
*  this code may duplicate columns, so a data reduction step may be needed. *;
proc sql noprint;
  create table work.new_columns_step4 as
  select new.sasref,
         new.table,
         new.column,
         case when new.label ne '' then new.label
              else ref.label
         end as label,
         new.order,
         new.type,
         new.length,
         ref.displayformat,
         ref.xmldatatype,
         ref.xmlcodelist,
         ref.core,
         ref.origin,
         ref.role,
         ref.term,
         ref.algorithm,
         ref.qualifiers,
         ref.standard,
         ref.standardversion,
         ref.standardref,
         ref.comment,
         1 as _cstfound from
  work.new_columns_step3 (where=(_cstfound=0)) new
   left join
  &_cstRefLib..&_cstRefColumnDS ref
  on substr(new.table,1,4) = 'SUPP' and substr(new.column,3) = substr(ref.column,3)
  where substr(new.table,1,4) = 'SUPP' and substr(ref.table,1,4) = 'SUPP'
  order by table, column ;
quit;

* Put all the columns back together again.  *;
data work.source_columns;
  set work.ref_columns
      work.new_columns_step1 (where=(_cstfound=1))
      work.new_columns_step2 (where=(_cstfound=1))
      work.new_columns_step3 (where=(_cstfound=1 or substr(table,1,4) ne 'SUPP'))
      work.new_columns_step4
  ;
run;

* Report unrecognized columns  *;
%let _cstDataRecords=0;
data work._cstProblems;
  set work.source_columns (where=(_cstfound=0)) end=last;

    %cstutil_resultsdskeep;
    attrib _cstSeqNo format=8. label="Sequence counter for result column"
           _cstMsgParm1 format=$char100. label="Message parameter value 1 (temp)"
           _cstMsgParm2 format=$char100. label="Message parameter value 2 (temp)"
           ;

    retain _cstSeqNo 0;
    if _n_=1 then _cstSeqNo=&_cstSeqCnt;

    keep _cstMsgParm1 _cstMsgParm2;

    * Set results data set attributes *;
    %cstutil_resultsdsattr;
    retain message resultseverity resultdetails '';

    srcdata = catx('.',sasref,table);
    resultid="CST0200";
    checkid="";
    _cstMsgParm1=catx(' ','No metadata found for column =',column);
    _cstMsgParm2='';
    resultseq=1;
    resultflag=1;
    _cst_rc=0;

    _cstSeqNo+1;
    seqno=_cstSeqNo;

    %if ^%symexist(_cstResultsDS) %then
    %do;
      put "No metadata found for " table= column=;
    %end;

    if last then
    do;
      call symputx('_cstSeqCnt',_cstSeqNo);
      call symputx('_cstDataRecords',_n_);
    end;
run;

%if &_cstDataRecords %then
%do;
  %cstutil_appendresultds( _cstErrorDS=work._cstProblems
                          ,_cstVersion=&_cstStandardVersion
                          ,_cstSource=CST
                          ,_cstStdRef=
                          );

%end;

* Write out final column-level source metadata  *;
proc sort data=work.source_columns (drop=_cstfound) out=&_cstTrgMetaLibrary..&_cstTrgColumnDS  (label="Source Column Metadata");
  by sasref table order;
run;

%if %symexist(_cstResultsDS) %then
%do;
  %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
  %cstutil_writeresult(
      _cstResultID=CST0074,
      _cstResultParm1=%sysfunc(pathname(&_cstTrgMetaLibrary)),
      _cstSeqNoParm=&_cstSeqCnt,
      _cstSrcdataParm=&_cstSrcData,
      _cstActualParm=&_cstTrgColumnDS
  );
  %******************************************************;
  %* Persist the results if specified in sasreferences  *;
  %******************************************************;
  %cstutil_saveresults();
%end;

* Clean-up  *;

%if %symexist(_cstResultsDS) %then
%do;
      proc datasets lib=work nolist;
  %if %eval(%index(&_cstResultsDS,.)>0) %then %do;
    %if %SYSFUNC(upcase(%scan(&_cstResultsDS,1,.)))=WORK %then %do;
        save %scan(&_cstResultsDS,2,.);
    %end;
  %end;
  %else %do;
        save &_cstResultsDS;
  %end;
      quit;
%end;
%else
%do;
  proc datasets lib=work kill nolist;
  quit;
%end;

%mend sendutil_createsrcmetafromsaslib;
