%* sendutil_createtumordataset                                                    *;
%*                                                                                *;
%* Creates a SEND TUMOR.XPT file and data set.                                    *;
%*                                                                                *;
%* This macro creates a TUMOR.XPT file and data set as specified for FDA          *;
%* submission in the SEND Version 3.0 Implementation Guide, Appendix C.           *;
%*                                                                                *;
%* NOTE: The following source domains are required: DM, TX, EX, DS, MI, and TF.   *;
%*                                                                                *;
%* @macvar studyRootPath Root path to the sample source study                     *;
%* @macvar _cstDebug Turns debugging on or off for the session                    *;
%* @macvar _cstGRoot Root path of the Toolkit Global Library                      *;
%* @macvar _cstResultsDS Results data set                                         *;
%* @macvar _cstSASRefs Run-time SASReferences data set derived in process setup   *;
%* @macvar _cstSeqCnt Results: Sequence number within _cstResultSeq               *;
%* @macvar _cstSrcData Results: Source entity being evaluated                     *;
%* @macvar _cstStandard Name of a standard registered to Toolkit                  *;
%* @macvar _cstStandardVersion Version of the standard referenced in _cstStandard *;
%*                                                                                *;
%* @param _cstSENDInputLibrary - required - The libname that points to the SEND   *;
%*            domains.                                                            *;
%* @param _cstSENDOutputDSFile - required - The libname.SAS data set name         *;
%*            (for example, study1.tumor) combination that specifies the data set *;
%*            name and library in which the content is written.                   *;
%* @param _cstSENDCreateXPT - optional - Generate a transport file.               *;
%*            Default: 1                                                          *;
%* @param _cstSENDOutputXPTFile - conditional - The full path and filename of the *;
%*            transport file to generate. If _cstSENDCreateXPT=1, a value is      *;
%*            required.                                                           *;
%*            Example:  c:\protocol_abc\study123\fda\tumor\rat.xpt                *;
%*                                                                                *;
%* @since 1.5                                                                     *;
%* @exposure external                                                             *;

%macro sendutil_createtumordataset(
    _cstSENDInputLibrary=,
    _cstSENDOutputDSFile=,
    _cstSENDCreateXPT=1,
    _cstSENDOutputXPTFile=
    ) / des='CST: Create FDA TUMOR.XPT from SEND Data';

  %local
    _cstReqDS
    _cstRN
    _cstrundt
    _cstVer
  ;

  %let _cstSeqCnt=0;
  %let _cstSrcData=&sysmacroname;


  %* Write information about this process to the results data set  *;
  %if %symexist(_cstResultsDS) %then
  %do;
    %cst_getRegisteredStandards(_cstOutputDS=work._cstStandards);

    data _null_;
      set work._cstStandards (where=(standard="CST-FRAMEWORK"));
      call symputx('_cstVer',strip(productrevision));
      call symputx('_cstrundt',put(datetime(),is8601dt.));
    run;
    proc datasets lib=work nolist;
      delete _cstStandards;
    quit;

    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STANDARD: &_cstStandard,_cstSeqNoParm=1,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STANDARDVERSION: &_cstStandardVersion,_cstSeqNoParm=2,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS DRIVER: CREATE_TUMORDATASET,_cstSeqNoParm=3,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS DATE: &_cstrundt,_cstSeqNoParm=4,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS TYPE: METADATA DERIVATION,_cstSeqNoParm=5,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS SASREFERENCES: &_cstsasrefs,_cstSeqNoParm=6,_cstSrcDataParm=&_cstSrcData);
    %if %symexist(studyRootPath) %then
      %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STUDYROOTPATH: &studyRootPath,_cstSeqNoParm=7,_cstSrcDataParm=&_cstSrcData);
    %else
      %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS STUDYROOTPATH: <not used>,_cstSeqNoParm=7,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS GLOBALLIBRARY: &_cstGRoot,_cstSeqNoParm=8,_cstSrcDataParm=&_cstSrcData);
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=PROCESS CSTVERSION: &_cstVer,_cstSeqNoParm=9,_cstSrcDataParm=&_cstSrcData);
    %let _cstSeqCnt=9;
  %end;

  %if &_cstDebug=1 %then
  %do;
    %put ************  ENTERING MACRO SENDUTIL_CREATETUMORDATASET  ************;
    %put _cstSENDInputLibrary =&_cstSENDInputLibrary;
    %put _cstSENDOutputDSFile =&_cstSENDOutputDSFile;
    %put _cstSENDCreateXPT    =&_cstSENDCreateXPT;
    %put _cstSENDOutputXPTFile=&_cstSENDOutputXPTFile;
    %put **********************************************************************;
  %end;

  %let _cst_rc=0;
  %*************************************************************;
  %*  Check for existence of needed parameters and data sets   *;
  %*************************************************************;
  %if &_cstSENDInputLibrary= %then
  %do;
    %put Note: The SEND input LIBRARY is required;
    %* Write information to the results data set about this run. *;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=TUMOR data set/xpt program cannot run - The SEND input library is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %if &_cstSENDOutputDSFile= %then
  %do;
    %put Note: The SEND data set output LIBRARY.<SASDATASET name> is required;
    %* Write information to the results data set about this run. *;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=TUMOR data set/xpt program cannot run - The SEND data set output library.name is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %if %length(&_cstSENDOutputXPTFile)=0 and &_cstSENDCreateXPT eq 1 %then
  %do;
    %put Note: The SEND transport file output PATHNAME and <SASXPTFILE name> is required;
    %* Write information to the results data set about this run. *;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=TUMOR data set/xpt program cannot run - The SEND transport output path and filename is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %if &_cst_rc=1 %then %goto exit_macro;

  %*******************************************;
  %*  Check existence of required data sets  *;
  %*  DM TX EX DS MI TF                      *;
  %*******************************************;
  %let _cstReqDS=&_cstSENDInputLibrary..dm;
  %if not %sysfunc(exist(&_cstReqDS)) %then
  %do;
    %put Warning: The SEND data set &_cstReqDS DOES NOT EXIST!;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=The SEND data set &_cstReqDS is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %let _cstReqDS=&_cstSENDInputLibrary..tx;
  %if not %sysfunc(exist(&_cstReqDS)) %then
  %do;
    %put Warning: The SEND data set &_cstReqDS DOES NOT EXIST!;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=The SEND data set &_cstReqDS is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %let _cstReqDS=&_cstSENDInputLibrary..ex;
  %if not %sysfunc(exist(&_cstReqDS)) %then
  %do;
    %put Warning: The SEND data set &_cstReqDS DOES NOT EXIST!;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=The SEND data set &_cstReqDS is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %let _cstReqDS=&_cstSENDInputLibrary..ds;
  %if not %sysfunc(exist(&_cstReqDS)) %then
  %do;
    %put Warning: The SEND data set &_cstReqDS DOES NOT EXIST!;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=The SEND data set &_cstReqDS is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %let _cstReqDS=&_cstSENDInputLibrary..mi;
  %if not %sysfunc(exist(&_cstReqDS)) %then
  %do;
    %put Warning: The SEND data set &_cstReqDS DOES NOT EXIST!;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=The SEND data set &_cstReqDS is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %let _cstReqDS=&_cstSENDInputLibrary..tf;
  %if not %sysfunc(exist(&_cstReqDS)) %then
  %do;
    %put Warning: The SEND data set &_cstReqDS DOES NOT EXIST!;
    %cstutil_writeresult(_cstResultID=CST0200,_cstResultParm1=The SEND data set &_cstReqDS is required,_cstSeqNoParm=1,_cstSrcDataParm=SENDUTIL_CREATETUMORDATASET);
    %let _cst_rc=1;
  %end;
  %****************************************************;
  %*  Required information is missing - Exit program  *;
  %****************************************************;
  %if &_cst_rc=1 %then %goto exit_macro;
  %else
  %do;
    %***************************************************************************************;
    %*  Generate random number for data set names to reduce filename conflicts during run  *;
    %***************************************************************************************;
    %cstutil_getRandomNumber(_cstVarname=_cstRN);
    %*******************************************************************************;
    %*  Prepare the DM and TX data to retrieve study and dosing group information  *;
    %*  Sort data for later merge to create the tumor data set                     *;
    %*******************************************************************************;
    proc sort data=&_cstSENDInputLibrary..dm out=work.dm&_cstRN(keep=studyid usubjid species sex setcd);
      by setcd;
    run;

    proc sort data=&_cstSENDInputLibrary..tx out=work.tx&_cstRN(keep=setcd txval);
      by setcd;
      where upcase(txparmcd)='SPGRPCD';
    run;

    data work.dg&_cstRN (keep=studyid dosegp usubjid species sex);
      merge work.dm&_cstRN work.tx&_cstRN;
      length dosegp 8.;
      by setcd;
      dosegp=put(txval,8.);
    run;

    proc sort data=work.dg&_cstRN;
      by studyid usubjid;
    run;

    %**************************************************************;
    %*  Prepare the EX and DS data to retrieve death information  *;
    %*  Sort data for later merge to create the tumor data set    *;
    %**************************************************************;
    proc sort data=&_cstSENDInputLibrary..ex out=work.ex&_cstRN(keep=studyid usubjid exstdtc);
      by studyid usubjid exstdtc;
      where exstdtc ne '';
    run;

    proc sort data=&_cstSENDInputLibrary..ds out=work.ds&_cstRN(keep=studyid usubjid dsstdtc dsdecod);
      by studyid usubjid;
    run;

    data work.di&_cstRN(keep=studyid usubjid dthsactm dthsacst);
      merge work.ex&_cstRN work.ds&_cstRN;
      by studyid usubjid;
      length dthsactm dthsacst 8.0;
      if length(dsstdtc) lt 19 or length(exstdtc) lt 19 then
      do;
       dsdt=input(substr(left(dsstdtc),1,10),YYMMDD10.);
       exdt=input(substr(left(exstdtc),1,10),YYMMDD10.);
      end;
      else
      do;
        dsdt=input(dsstdtc,E8601DT.);
        exdt=input(exstdtc,E8601DT.);
      end;
      dthsactm=(dsdt-exdt)+1;
      dthsacst=put(put(dsdecod, $DTHCAT.),8.);
    run;

    %********************************************************************************;
    %*  Prepare the MI and TF data to retrieve organ and tumor related information  *;
    %*  Sort data for later merge to create the tumor data set                      *;
    %********************************************************************************;
    proc sort data=&_cstSENDInputLibrary..mi out=work.mi&_cstRN(keep=studyid usubjid mistat mispcufl mispid);
      by studyid usubjid;
    run;

    data work.ae&_cstRN(keep=studyid usubjid animlexm) work.m&_cstRN(keep=studyid usubjid mispid organexm);
      set work.mi&_cstRN;
      by studyid usubjid;
      length animlexm organexm 8.;
      retain animlexm;

      if first.usubjid then animlexm=0;
      if upcase(mistat) ne 'NOT DONE' then animlexm=1;

      if strip(mistat)='' and strip(mispcufl)='' then organexm=1;
        else if upcase(mispcufl)='N' then organexm=2;
        else if upcase(mistat)='NOT DONE' and strip(mispcufl)='' then organexm=3;

      output work.m&_cstRN;
      if last.usubjid then output work.ae&_cstRN;
    run;

    data work.mi&_cstRN;
      merge work.ae&_cstRN work.m&_cstRN;
      by studyid usubjid;
    run;

    %*********************************************************;
    %*  Reduce size of data set to reflect multiple records  *;
    %*  for animals with tumor and single records for those  *;
    %*  without tumor                                        *;
    %*********************************************************;
    proc sort data=work.mi&_cstRN nodupkey;
      by studyid usubjid  organexm mispid;
    run;

    data work.mi&_cstRN;
      set work.mi&_cstRN;
      by studyid usubjid;
      if not(first.usubjid and last.usubjid) and mispid="" then delete;
    run; 
 
    proc sql;
      create table work.ms&_cstRN as select a.studyid, a.usubjid, animlexm, organexm, tfstresc, tfspec, tfdetect, tfrescat, tfdthrel
      from  work.mi&_cstRN as a LEFT JOIN &_cstSENDInputLibrary..tf as b
      on a.studyid=b.studyid and a.usubjid=b.usubjid and a.mispid=b.tfspid;
    quit;

    data work.ms&_cstRN(drop=tfdthrel tfrescat);
      set work.ms&_cstRN(rename=(tfstresc=tumornam tfspec=organnam tfdetect=detecttm));
      length deathcau malignst 8.;
      deathcau=put(put(tfdthrel,$DTHCAU.),8.);
      malignst=put(put(tfrescat,$MALIG.),8.);
    run;

    %****************************************************************************;
    %*  Create working copy tumor data set by merging files created previously  *;
    %*  Sort data for later merge to create the tumor data set                  *;
    %****************************************************************************;
    data work.tu&_cstRN;
      attrib
        studynum label='Study number'
        animlnum label='Animal number'
        species label='Animal species'
        sex label='Sex'
        dosegp format=8. label='Dose group'
        dthsactm label='Time in days to death or sacrifice'
        dthsacst label='Death or sacrifice status'
        animlexm format=8. label='Animal microscopic examination code'
        tumorcod format=$8. label='Tumor type code'
        tumornam label='Tumor name'
        organcod format=$8. label='Organ/tissue code'
        organnam label='Organ/tissue name'
        detecttm format=8. label='Time in days to detection of tumor'
        malignst format=8. label='Malignancy status'
        deathcau format=8. label='Cause of death'
        organexm format=8. label='Organ/tissue microscopic examination code';

      merge work.ms&_cstRN work.di&_cstRN work.dg&_cstRN;
      by studyid usubjid;
      studynum=studyid;
      animlnum=usubjid;
      organcod=put(organnam,$ORGANCD.);
      tumorcod=put(tumornam,$TUMORCD.);
      species=put(species,$SPECIES.);
      if organexm in (2,3) then
      do;
        malignst=.;
        deathcau=.;
        tumornam='';
        tumorcod='';
      end;
      drop studyid usubjid;
    run;

    %**********************************;
    %*  Produce final tumor data set  *;
    %**********************************;
    proc sort data=work.tu&_cstRN out=&_cstSENDOutputDSFile;
      by studynum animlnum tumorcod;
    run;

    %if (&syserr le 4 and %symexist(_cstResultsDS)) %then
    %do;
      %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
      %cstutil_writeresult(
          _cstResultID=CST0102,
          _cstResultParm1=&_cstSENDOutputDSFile,
          _cstSeqNoParm=&_cstSeqCnt,
          _cstSrcdataParm=&_cstSrcData,
          _cstActualParm=
      );
    %end;

    %**************************************;
    %*  Produce final tumor data set XPT  *;
    %**************************************;
    %if &_cstSENDCreateXPT eq 1 %then
    %do;
      filename tf&_cstRN "&_cstSENDOutputXPTFile";
      proc cport data=work.tu&_cstRN file=tf&_cstRN;
      run;
      %if (&syserr le 4 and %symexist(_cstResultsDS)) %then
      %do;
        %let _cstSeqCnt=%eval(&_cstSeqCnt+1);
        %cstutil_writeresult(
            _cstResultID=CST0102,
            _cstResultParm1=&_cstSENDOutputXPTFile,
            _cstSeqNoParm=&_cstSeqCnt,
            _cstSrcdataParm=&_cstSrcData,
            _cstActualParm=
        );
      %end;
      filename tf&_cstRN;
    %end;

    %*****************************;
    %*  Cleanup temporary files  *;
    %*****************************;
    %if &_cstDebug ne 1 %then
    %do;
      proc datasets lib=work nolist;
        delete dm&_cstRN/memtype=data;
        delete tx&_cstRN/memtype=data;
        delete dg&_cstRN/memtype=data;
        delete ex&_cstRN/memtype=data;
        delete ds&_cstRN/memtype=data;
        delete di&_cstRN/memtype=data;
        delete mi&_cstRN/memtype=data;
        delete ae&_cstRN/memtype=data;
        delete m&_cstRN/memtype=data;
        delete ms&_cstRN/memtype=data;
        delete tu&_cstRN/memtype=data;
      quit;
    %end;
  %end;

%exit_macro:

%if &_cstDebug=1 %then
%do;
  %put ************  LEAVING MACRO SENDUTIL_CREATETUMORDATASET  ************;
  %put _cstSENDInputLibrary =&_cstSENDInputLibrary;
  %put _cstSENDOutputDSFile =&_cstSENDOutputDSFile;
  %put _cstSENDCreateXPT    =&_cstSENDCreateXPT;
  %put _cstSENDOutputXPTFile=&_cstSENDOutputXPTFile;
  %put _cst_rc              =&_cst_rc;
  %put *********************************************************************;
%end;

%if %symexist(_cstResultsDS) %then
%do;
  %******************************************************;
  %* Persist the results if specified in sasreferences  *;
  %******************************************************;
  %cstutil_saveresults();
%end;


%mend sendutil_createtumordataset;
